#!/usr/bin/env python3
# 
# Copyright 2020-2021 by Andreas Schmidt
# All rights reserved.
# This file is part of the trackfs project
# and licensed under the terms of the GNU Lesser General Public License v3.0.
# See https://github.com/andresch/trackfs for details.
#

#
# This module provides the cuesheet parsing functionality for trackfs.
# While it specifies a fully fledged parser for cuesheets, only those
# elements needed in the context for trackfs (which is the track information) 
# get extracted and exposed.
# 
# This implementation uses original spec of the cuesheet format as found on archive.org
# https://web.archive.org/web/20070614044112/http://www.goldenhawk.com/download/cdrwin.pdf
#

import math
from dataclasses import dataclass
from lark import Lark, Transformer

import logging

from typing import List

log = logging.getLogger(__name__)

# Lark cue-sheet grammar according to original spec
_CUE_LARK_GRAMMAR = r"""
   cue_sheet      : disc_entries tracks
   disc_entries   : disc_entry* 
   tracks         : track*
    
   ?disc_entry : catalog 
      | comment
      | performer
      | songwriter
      | title
      | file

   track          : "TRACK" INT TRACKTYPE track_entries
   track_entries  : ( track_entry )*
   ?track_entry   : isrc
      | flags
      | pregap
      | index
      | postgap
      | comment
      | performer
      | songwriter
      | title
      | file
      
   catalog        : "CATALOG" UPC_EAN
   comment        : "REM" REST_OF_LINE
   performer      : "PERFORMER" STRING
   songwriter     : "SONGWRITER" STRING
   title          : "TITLE" STRING
   file           : "FILE" STRING FILETYPE
   isrc           : "ISRC" ISRC
   flags          : "FLAGS" FLAG+
   index          : "INDEX" INDEX mmssff
   pregap         : "PREGAP" mmssff
   postgap        : "POSTGAP" mmssff
   mmssff         : TIME_ELEM ":" TIME_ELEM ":" TIME_ELEM
   
   TIME_ELEM      : DIGIT ~ 2
   INDEX          : DIGIT ~ 1..2
   UPC_EAN        : DIGIT ~ 12..13
   REST_OF_LINE   : /[^\n]*/ NEWLINE
   ISRC           : LETTER LETTER (DIGIT|LETTER) ~ 10
   
   STRING : ("\"" /.*?/ "\"") | /[^ \n]+/
   FILETYPE : "BINARY"  // Intel binary file (LSBF). Use for data files.
      | "MOTOROLA"      // Motorola binary file (MSBF). Use for data files.
      | "AIFF"          // Audio AIFF file (44.1KHz 16-bit stereo)
      | "WAVE"          // Audio WAVE file (44.1KHz 16-bit stereo)
      | "MP3"           // Audio MP3 file (44.1KHz 16-bit stereo)

   TRACKTYPE: "AUDIO"   // Audio/Music (2352)
      | "CDG"           // Karaoke CD+G (2448)
      | "MODE1/2048"    // CD-ROM Mode1 Data (cooked)
      | "MODE1/2352"    // CD-ROM Mode1 Data (raw)
      | "MODE2/2336"    // CD-ROM XA Mode2 Data
      | "MODE2/2352"    // CD-ROM XA Mode2 Data
      | "CDI/2336"      // CD-I Mode2 Data
      | "CDI/2352"      // CD-I Mode2 Data

   FLAG: "DCP"       // Digital copy permitted
      | "4CH"        // Four channel audio
      | "PRE"        // Pre-emphasis enabled (audio tracks only)
      | "SCMS"       // Serial Copy Management System
      | "DATA"       // set for data files

   %import common.NEWLINE
   %import common.INT
   %import common.LETTER
   %import common.DIGIT
   %import common.ESCAPED_STRING
   %import common.LETTER
   %import common.WS
   %ignore WS"""

_CUE_LARK_PARSER = Lark(_CUE_LARK_GRAMMAR, start='cue_sheet')


@dataclass
class CueSheet:
    """Meta data from a parsed cue sheet
   
    """
    tracks: List['Track']
    albumartists: List[str] = None
    album: str = None
    composers: List[str] = None
    catalog: str = None
    discid: str = None
    year: str = None
    discnumber: str = None
    totaldiscs: str = None

    def calc_track_times(self, disc_duration):
        for i in range(0, len(self.tracks) - 1):
            curr = self.tracks[i]
            curr.end = self.tracks[i + 1].start
            curr.duration = curr.end - curr.start
        last = self.tracks[-1];
        last.end = Time.create(disc_duration)
        return self

@dataclass
class Track:
    """All information extracted for a track from a cue sheet

    Attributes
    ----------
    num         :  int
                   The track number
    type        :  string
                   The track type
    start       :  Time
                   The start timestamp of the track
    end         :  Time
                   The end timestamp of the track
    artists     :  list[string]
                   The artists performing the track
    composers   :  list[string]
                   The artists who have written the track

    The attributes end and duration don't originate directly from the the track elements
    in the cue sheet. They get set by calling a CueSheet's `calc_track_times` method
    once the whole cue sheet got parsed.
    """

    num: int
    type: str
    artists: List[str] = None
    composers: List[str] = None
    title: List[str] = None
    isrc: str = None
    start: 'Time' = None
    end: 'Time' = None
    duration: 'Time' = None


@dataclass(frozen=True)
class Time:
    """ Timestamp / duration information with CD frame accuracy

    Attributes
    ----------
    mm      :   int
                minutes
    ss      :   int
                seconds
    ff      :   int
                sub-second CD frames (75 frames per second)


    The constructor supports various initializations:
    - <empty>      :  mm = ss == ff = 0
    - float        : interpreted as seconds (and fractions of)
    - int-triple   : interpreted as (mm, ss, ff
    - 6-charstring : interpreted as "mmssff"-string
    - three int    : interpreted as mm, ss, ff

    The class supports basic math (+,-)
    The string representation is "mmsscc"; `flac_time' returns the string 
    representation than FLAC expects ("mm:ss.cc")
    """

    mm: int
    ss: int
    ff: int

    @classmethod
    def create(cls, *args):
        if len(args) == 0:
            return cls(0, 0, 0)
        elif isinstance(args[0], float):
            # value as seconds with faction
            (sec, frac) = divmod(args[0], 1)
            (mm, ss) = divmod(int(sec), 60)
            return cls(mm, ss, int(frac * 75))
        elif isinstance(args[0], tuple):
            # value is tuple (minutes, seconds, frames)
            (mm, ss, ff) = args[0]
            return cls(mm, ss, ff)
        elif isinstance(args[0], str):
            # value is string template "Time"
            ts = args[0]
            assert len(ts) == 6
            return cls(int(ts[0:2]), int(ts[2:4]), int(ts[4:6]))
        elif isinstance(args[0], int):
            # value is three ints: mm, ss, ff
            assert len(args) == 3
            return cls(*args)

    def seconds(self):
        return (60.0 * self.mm) + self.ss + (self.ff / 100.0)

    def flac_time(self):
        return f'{self.mm:02d}:{self.ss:02d}.{int(100.0 / 75.0 * self.ff):02d}'

    def __repr__(self):
        return '%02d%02d%02d' % (self.mm, self.ss, self.ff)

    def __add__(self, other):
        (cs, ff) = divmod(self.ff + other.ff, 75)
        (cm, ss) = divmod(self.ss + other.ss + cs, 60)
        return Time(self.mm + other.mm + cm, ss, ff)

    def __sub__(self, other):
        ff = self.ff - other.ff
        ss = self.ss - other.ss - (0 if ff > 0 else 1)
        return Time(
            self.mm - other.mm - (0 if ss > 0 else 1),
            ss if ss >= 0 else ss + 60,
            ff if ff >= 0 else ff + 75
        )


class _CueTransformer(Transformer):
    """Transforms the Lark-parser-tree in a CueSheet"""

    @staticmethod
    def _extend__list_arg(args, name, value):
        # most rippers use ";" as delimiter for multiple values inside a
        # single entry rather than having multiple entries
        values = [p.strip() for p in value.split(";")]
        old_value = args.get(name, None)
        if old_value is None:
            args[name] = values
        else:
            args[name] = old_value + values

    @staticmethod
    def _map_to_arg(args, mappings, name, value) -> bool:
        mapping = mappings.get(name, None)
        if mapping is None:
            return False
        (mapped_name, is_list) = mapping
        if is_list:
            _CueTransformer._extend__list_arg(args, mapped_name, value)
        else:
            args[mapped_name]=value
        return True

    ALBUM_MAPPINGS = {
        'performer': ("albumartists", True),
        'songwriter': ("composers", True),
        'title': ("album", False),
        'catalog': ("catalog", False)
    }

    COMMENT_MAPPINGS = {
        'discid': ("discid", False),
        'date': ("year", False),
        'discnumber': ("discnumber", False),
        'totaldiscs': ("totaldiscs", False)
    }

    def cue_sheet(self, subtrees):
        args = {'tracks': subtrees[1].children}
        for entry in subtrees[0].children:
            (name, value) = (entry.data.lower(), entry.children[0])
            if _CueTransformer._map_to_arg(args, self.ALBUM_MAPPINGS, name, value):
                pass
            elif name == 'comment':
                # extract known key-value comments
                splits = value.split(' ', 1)
                if len(splits) > 1:
                    _CueTransformer._map_to_arg(args, self.COMMENT_MAPPINGS, splits[0].lower(), splits[1])
        return CueSheet(**args)

    TRACK_MAPPINGS = {
        'performer': ("artists", True),
        'songwriter': ("composers", True),
        'title': ("title", False),
        'isrc': ("isrc", False)
    }

    def track(self, subtrees):
        args = {'num': subtrees[0], 'type': subtrees[1]}
        for entry in subtrees[2].children:
            (name, value) = (entry.data.lower(), entry.children[0])
            if _CueTransformer._map_to_arg(args, self.TRACK_MAPPINGS, name, value):
                pass
            elif name == 'index':
                idx = entry.children
                # we're only interested in index 1
                if idx[0] == 1:
                    args['start'] = idx[1]
        return Track(**args)

    def mmssff(self, elems):
        return Time(elems[0], elems[1], elems[2])

    def REST_OF_LINE(self, comment_line):
        return comment_line[1:-1] if len(comment_line) > 0 else ""

    def STRING(self, string):
        return string[1:-1] if string[0] == '"' else str(string)

    UPC_EAN = str
    ISRC = str
    TRACKTYPE = str
    INT = int
    INDEX = int
    TIME_ELEM = int


def parse(cuesheet, disc_total_length):
    return _CueTransformer(visit_tokens=True
                           ).transform(_CUE_LARK_PARSER.parse(cuesheet)
                                       ).calc_track_times(disc_total_length)


if __name__ == '__main__':
    test_cue = r"""REM DISCID A10A2E0D
    PERFORMER "Zaz"
    TITLE "Paris"
    CATALOG 5054196339524
    REM DATE 2014
    REM DISCNUMBER 1
    REM TOTALDISCS 1
    REM COMMENT "CUERipper v2.1.4 Copyright (C) 2008-12 Grigory Chudov"
    FILE "Zaz - Paris.flac" WAVE
      TRACK 01 AUDIO
        PERFORMER Zaz
        PERFORMER Zazo
        TITLE "Paris sera toujours Paris"
        ISRC FR2PY1403200
        INDEX 01 00:00:00
      TRACK 02 AUDIO
        PERFORMER "Zaz; Zazo"
        TITLE "Sous le ciel de Paris"
        ISRC FR2PY1403250
        INDEX 01 02:58:68"""

    print(
        _CueTransformer(
            visit_tokens=True
        ).transform(
            _CUE_LARK_PARSER.parse(test_cue)
        )
    )



#!/usr/bin/env python3
# 
# Copyright 2020-2021 by Andreas Schmidt
# All rights reserved.
# This file is part of the trackfs project
# and licensed under the terms of the GNU Lesser General Public License v3.0.
# See https://github.com/andresch/trackfs for details.
#

#
# This module provides mapping of names between the virtual trackfs 
# filesystem and the underlying root filesystem, esp. the naming of
# the individual track-files
#

import os
import re
import string
import unicodedata

from dataclasses import dataclass, field
from functools import cached_property

from . import cuesheet
from . import albuminfo

import logging
log = logging.getLogger(__name__)

DEFAULT_TRACK_SEPARATOR     : str   = '.#-#.'
DEFAULT_MAX_TITLE_LEN       : int   = 20
DEFAULT_FLAC_EXTENSION      : str   = '.flac'
DEFAULT_VALID_CHARS         : str   = "-_() " + string.ascii_letters + string.digits
DEFAULT_KEEP_PATH           : bool  = False

@dataclass(frozen=True)
class Factory():
    '''manages the configuration options for the virtual fuse paths'''
    
    track_separator         : str   = DEFAULT_TRACK_SEPARATOR
    max_title_len           : int   = DEFAULT_MAX_TITLE_LEN
    flac_extension          : str   = DEFAULT_FLAC_EXTENSION
    valid_filename_chars    : str   = DEFAULT_VALID_CHARS
    keep_flac               : bool  = DEFAULT_KEEP_PATH
    
    @cached_property
    def track_file_regex(self):
        (separator_rex, extension_rex) = [ s.replace('.','\\.') for s in [self.track_separator, self.flac_extension] ]
        flac_cue_rex = (
            '^(?P<basename>.*)'+separator_rex
            + '(?P<num>\\d+)(?P<title>(\\.[^\\.]{,'+str(self.max_title_len)
            + '}?)?)\\.(?P<start>\\d{6})-(?P<end>\d{6})'
            + '(?P<extension>'+extension_rex+')$'
        )
        log.debug("Factory.track_file_regex: "+flac_cue_rex)
        return re.compile(flac_cue_rex)

    def from_vpath(self, path):
        """Construct a FusePath instance from a given virtual path"""
        match = self.track_file_regex.match(path)
        if match is None: 
            log.debug(f'no track file in "{path}"');
            (root, ext) = os.path.splitext(path)
            return FusePath(root, ext, _factory=self)
        log.debug(f'track file in "{path}"');
        title = match['title'].lstrip()
        return FusePath(
            match['basename'], match['extension'], True,
            int(match['num']), title,
            cuesheet.Time.create(match['start']), cuesheet.Time.create(match['end']),
            self
        )
        
    def from_track(self, source_root, extension, track):
        return FusePath(
            source_root, extension, True,
            track.num, track.title, track.start, track.end,
            self
        )      
        
_DEFAULT_FACTORY = Factory()

@dataclass(frozen=True)
class FusePath():
    ''' represents an entry in the virtual trackfs filesystem'''
    source_root             : str
    extension               : str
    is_track                : bool          = False
    num                     : int           = None
    title                   : str           = None
    start                   : cuesheet.Time = None    
    end                     : cuesheet.Time = None
    _factory                : Factory       = _DEFAULT_FACTORY
        
    @property
    def track_separator(self): return self._factory.track_separator
    @property
    def max_title_len(self): return self._factory.max_title_len
    @property
    def flac_extension(self): return self._factory.flac_extension
    @property
    def valid_filename_chars(self): return self._factory.valid_filename_chars
    @property
    def track_file_regex(self): return self._factory.track_file_regex
    @property
    def keep_flac(self): return self._factory.keep_flac

    @cached_property
    def source(self):
        return self.source_root + self.extension
    
    @property
    def title_fragment(self):
        '''the fragment of a track's title that goes into a vpath'''
        if self.title is None or len(self.title) == 0: 
            return ""
        else:
            clean_title = unicodedata.normalize('NFKD', self.title)[:self.max_title_len]
            return "."+''.join(c if c in self.valid_filename_chars else "_" for c in clean_title)
        
    @property
    def vpath(self):
        if(self.is_track): 
            return (
                f'{self.source_root}{self.track_separator}{self.num:03d}'
                f'{self.title_fragment}.{self.start}-{self.end}{self.extension}'
            )
        else:  
            return self.source
            
    def dirname(self):
        return os.path.dirname(self.source_root)
    
    def readdir(self):
        entries = ['.', '..']
        for filename in os.listdir(self.source):
            (basename, extension) = os.path.splitext(filename)
            if( extension == self.flac_extension ):
                trx = albuminfo.get(os.path.join(self.source, filename)).tracks()
                if trx:
                    if self.keep_flac:
                        entries.append(filename)
                    for t in trx:
                        entries.append( 
                            self._factory.from_track(basename, extension, t).vpath
                        )
                else:
                   entries.append(filename)
            else:
                entries.append(filename)
        log.debug(f'vdir entries:{entries}')
        return entries

    def for_other_track(self, num: int, title: str, start: cuesheet.Time, end: cuesheet.Time):
        '''Construct fusepath entry for another track of the same FLAC+CUE file'''
        return FusePath(
            self.source_root, self.extension, True,
            num, title, start, end,
            self._factory
        )


#!/usr/bin/env python3
# 
# Copyright 2020-2021 by Andreas Schmidt
# All rights reserved.
# This file is part of the trackfs project
# and licensed under the terms of the GNU Lesser General Public License v3.0.
# See https://github.com/andresch/trackfs for details.
#
# This file is derived work of the FLACCue project.
# See https://github.com/acenko/FLACCue for details
#

from __future__ import print_function, absolute_import, division

import os
from dataclasses import dataclass
from threading import RLock

from fuse import Operations

from . import fusepath
from . import albuminfo
from .flactracks import TrackManager

import logging

log = logging.getLogger(__name__)


@dataclass
class OpenFileInfo:
    position: int = 0
    lock: RLock = RLock()


class TrackFSOps(Operations):

    def __init__(self,
                 root,
                 keep_flac=fusepath.DEFAULT_KEEP_PATH,
                 separator=fusepath.DEFAULT_TRACK_SEPARATOR,
                 flac_extension=fusepath.DEFAULT_FLAC_EXTENSION,
                 title_length=fusepath.DEFAULT_MAX_TITLE_LEN,
                 tags_ignored=albuminfo.DEFAULT_IGNORE_TAGS_REX
                 ):
        self.root = os.path.realpath(root)
        self.keep_flac = keep_flac
        self.tracks = TrackManager()
        self._open_files = {}
        self._fusepath_factory = fusepath.Factory(
            track_separator=separator,
            max_title_len=title_length,
            flac_extension=flac_extension,
            keep_flac=keep_flac
        )
        # TODO: avoid global init function
        albuminfo.init(tags_ignored)

    def __call__(self, op, path, *args):
        return super(TrackFSOps, self).__call__(op, self.root + path, *args)

    def _fusepath(self, path):
        return self._fusepath_factory.from_vpath(path)

    def getattr(self, path, fh=None):
        log.info(f"getattr for ({path}) [{fh}]")
        fp = self._fusepath(path)
        log.debug(fp)
        st = os.lstat(fp.source)
        result = dict((key, getattr(st, key)) for key in (
            'st_atime', 'st_ctime', 'st_gid', 'st_mode', 'st_mtime',
            'st_nlink', 'st_size', 'st_uid'))

        if (fp.is_track):
            # If it's one of the FlacTrackFS track paths, 
            # we need to adjust the file size to be roughly
            # appropriate for the individual track.
            result['st_size'] = self.tracks.estimate_track_file_size(path, fp)

        return result

    getxattr = None

    listxattr = None

    def open(self, path, flags, *args, **pargs):
        log.info(f'open file "{path}"')
        # We don't want FlacTrackFS messing with actual data.
        # Only allow Read-Only access.
        if (flags | os.O_RDONLY) == 0:
            raise ValueError('Can only open files read-only.')
        fp = self._fusepath(path)
        if fp.is_track:
            path = self.tracks.prepare_track(path, fp)
        log.debug(f'file to open = "{path}"')
        fh = os.open(path, flags, *args, **pargs)
        self._open_files[fh] = OpenFileInfo()
        log.debug(f'opened file file to open = "{path}" with fh [{fh}]')
        return fh

    def read(self, path, size, offset, fh):
        log.info(f"read from [{fh}] {offset} until {offset + size}")
        open_file_info = self._open_files[fh]
        # make sure that only one concurrent read per file handle is possible
        with open_file_info.lock:
            if open_file_info.position != offset:
                log.debug(f"out of band read; seek file to offset {offset}")
                os.lseek(fh, offset, 0)
            else:
                # we do preload-checks only on consecutive reads
                fp = self._fusepath(path)
                if fp.is_track:
                    self.tracks.check_next_track(path, fp, offset)
            open_file_info.position = offset + size
            return os.read(fh, size)

    def release(self, path, fh):
        log.info(f'release [{fh}] ({path})')
        del self._open_files[fh]
        fp = self._fusepath(path)
        if fp.is_track:
            self.tracks.release_track(path, fp)
        return os.close(fh)

    def readdir(self, path, fh):
        log.info(f'readdir [{fh}] ({path})')
        return self._fusepath(path).readdir()

    def readlink(self, path, *args, **pargs):
        log.info(f'readlink ({path})')
        path = self._fusepath(path).source
        return os.readlink(path, *args, **pargs)

    def statfs(self, path):
        log.info(f'statfs ({path})')
        path = self._fusepath(path).source
        stv = os.statvfs(path)
        return dict((key, getattr(stv, key)) for key in (
            'f_bavail', 'f_bfree', 'f_blocks', 'f_bsize', 'f_favail',
            'f_ffree', 'f_files', 'f_flag', 'f_frsize', 'f_namemax')
                    )



# 
# Copyright 2020-2021 by Andreas Schmidt
# All rights reserved.
# This file is part of the trackfs project
# and licensed under the terms of the GNU Lesser General Public License v3.0.
# See https://github.com/andresch/trackfs for details.
#

import setuptools

def slurp(fn):
    with open(fn, "r", encoding="utf-8") as fh:
        return fh.read()


setuptools.setup(
    name="trackfs",
    version=slurp("VERSION"),
    author="Andreas Schmidt",
    author_email="author@example.com",
    description="A read-only FUSE filesystem that splits FLAC+CUE files into individual FLAC files per track",
    long_description=slurp("README.py.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/andresch/trackfs",
    packages=setuptools.find_packages(),
    install_requires=[
       "mutagen", "fusepy", "Lark", "chardet"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: POSIX",
    ],
    python_requires='>=3.8',
    entry_points={
      'console_scripts': [
         'trackfs=trackfs.__init__:main'
      ],
    }
)



#!/usr/bin/env python3
# 
# Copyright 2020-2021 by Andreas Schmidt
# All rights reserved.
# This file is part of the trackfs project
# and licensed under the terms of the GNU Lesser General Public License v3.0.
# See https://github.com/andresch/trackfs for details.
#

from __future__ import print_function, absolute_import, division

from .fuseops import TrackFSOps


def main(foreground=True, allow_other=True):
    import os
    import sys
    import argparse

    from fuse import FUSE
    from . import fusepath

    import logging
    log = logging.getLogger(__name__)

    parser = argparse.ArgumentParser(
        description='''A FUSE filesystem for extracting individual tracks from FLAC+CUE files.

        Maps a directory to a mount point while replacing all FLAC+CUE files (with 
        embedded cue sheets) with multiple FLAC files for the individual tracks''')
    parser.add_argument(
        '-s', '--separator', dest='separator', default=fusepath.DEFAULT_TRACK_SEPARATOR,
        help=(
            f'The separator used inside the name of the track-files. '
            f'Must never occur in regular filenames (default: "{fusepath.DEFAULT_TRACK_SEPARATOR}")'
        )
    )
    parser.add_argument(
        '-i', '--ignore-tags', dest='ignore', default='CUE_TRACK.*|COMMENT',
        help=(
            'A regular expression for tags in the FLAC file that will not be '
            'copied to the track FLACs (default: "CUE_TRACK.*|COMMENT")'
        )
    )
    parser.add_argument(
        '-e', '--extension', dest='extension', default=fusepath.DEFAULT_FLAC_EXTENSION,
        help=f'The file extension of FLAC files (default: "{fusepath.DEFAULT_FLAC_EXTENSION}")'
    )
    parser.add_argument(
        '-k', '--keep-flac-cue', dest='keep', action='store_true',
        help='Keep the source FLAC+CUE file in the mapped filesystem'
    )
    parser.add_argument(
        '-t', '--title-length', dest='title_length', default=fusepath.DEFAULT_MAX_TITLE_LEN,
        help=f'Nr. of characters of the track title in filename of track (default: {fusepath.DEFAULT_MAX_TITLE_LEN})'
    )
    parser.add_argument(
        '--root-allowed', dest='rootok', action='store_true',
        help=(
            'Allow running as with root permissions; Neither necessary nor recommended. '
            'Use only when you know what you are doing'
        )
    )
    parser.add_argument(
        '-v', '--verbose', dest='verbose', action='store_true',
        help='Activate info-level logging'
    )
    parser.add_argument(
        '-d', '--debug', dest='debug', action='store_true',
        help='Activate debug-level logging'
    )
    parser.add_argument(
        'root',
        help='The root of the directory tree to be mapped'
    )
    parser.add_argument(
        'mount',
        help='The mount point for the mapped directory tree'
    )
    args = parser.parse_args()

    log_fmt = '{levelname:6s}[{threadName:15s}]({module:10}) {message}'
    if args.debug:
        logging.basicConfig(format=log_fmt, style='{', level=logging.DEBUG )
        log.setLevel(logging.DEBUG)
    elif args.verbose:
        logging.basicConfig(format=log_fmt, style='{', level=logging.INFO )
        log.setLevel(logging.INFO)

    if os.geteuid() == 0 and not args.rootok:
        print(
            f'''By default {os.path.basename(sys.argv[0])} don't allow to run with root permissions. 
     
If you are absolutely sure that that's what you want, use the option "--root-allowed"''',
            file=sys.stderr
        )
        exit(1)

    trackfs = TrackFSOps(
        args.root,
        keep_flac=args.keep, separator=args.separator, flac_extension=args.extension,
        title_length=int(args.title_length), tags_ignored=args.ignore
    )

    fuse = FUSE(trackfs, args.mount, foreground=foreground, allow_other=allow_other)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# 
# Copyright 2020-2021 by Andreas Schmidt
# All rights reserved.
# This file is part of the trackfs project
# and licensed under the terms of the GNU Lesser General Public License v3.0.
# See https://github.com/andresch/trackfs for details.
#

import re
import os
import shlex
from functools import lru_cache, cached_property

from mutagen.flac import FLAC
import chardet

from . import cuesheet

import logging

log = logging.getLogger(__name__)

DEFAULT_IGNORE_TAGS_REX = re.compile('CUE_TRACK.*|COMMENT')


class AlbumInfo():
    IGNORE_TAGS_REX = DEFAULT_IGNORE_TAGS_REX

    def __init__(self, path):
        self.path: str = path

    @cached_property
    def meta(self) -> FLAC:
        return FLAC(self.path)

    def _cue_from_external_file(self):
        (base, ext) = os.path.splitext(self.path)
        cue_path = base + ".cue"
        if not os.path.exists(cue_path):
            return None
        log.debug(f"found accompanying cue sheet")
        with open(cue_path, "rb") as fh:
            cue_bytes = fh.read()
        cue_str = cue_bytes.decode(chardet.detect(cue_bytes)['encoding'])
        log.debug(f"cue-sheet:\n{cue_str}")
        return cue_str

    @cached_property
    def cue(self):
        meta = self.meta
        raw_cue = meta.tags.get('CUESHEET', [])
        if len(raw_cue) == 0:
            log.debug(f"regular flac file without cue sheet")
            raw_cue = self._cue_from_external_file()
            if raw_cue is None:
                return None
        else:
            raw_cue =raw_cue[0]
        log.debug(f"raw cue sheet from FLAC file:\n{raw_cue}")
        result = cuesheet.parse(raw_cue, meta.info.length)
        log.debug(f"parsed cue sheet from FLAC file:\n{result}")
        return result

    def tracks(self):
        return self.cue.tracks if self.cue is not None else None

    def track(self, num):
        trx = self.tracks()
        if trx is None:
            return None
        t = trx[num - 1]
        if t.num == num:
            return t
        for t in trx:
            if t.num == num:
                return t
        return None

    def _cue_album_tags(self):
        cue = self.cue
        tags = {}
        if cue is not None:
            if cue.albumartists: tags['ALBUMARTIST'] = cue.albumartists
            if cue.composers: tags['COMPOSER'] = cue.composers
            if cue.album: tags['ALBUM'] = [cue.album]
            if cue.catalog: tags['CATALOG'] = [cue.catalog]
            if cue.discid: tags['DISCID'] = [cue.discid]
            if cue.year: tags['YEAR'] = [cue.year]
            if cue.discnumber: tags['DISCNUMBER'] = [cue.discnumber]
            if cue.totaldiscs: tags['TOTALDISCS'] = [cue.totaldiscs]

        return tags

    def _album_tags(self):
        meta = self.meta
        tags = {}
        for (k, v) in meta.tags:
            k = k.upper()
            # skip multi-line tags
            if len(v.splitlines()) != 1:
                continue
            # skip _IGNORE_TAGS
            if AlbumInfo.IGNORE_TAGS_REX.match(k):
                continue
            if k not in tags: tags[k] = []
            tags[k].append(v)

        # make sure ALBUMARTIST and ALBUM are set
        # in case ARTIST and TITLE have been used instead
        if 'ALBUMARTIST' not in tags and 'ARTIST' in tags:
            tags['ALBUMARTIST'] = tags['ARTIST']
        if 'ALBUM' not in tags and 'TITLE' in tags:
            tags['ALBUM'] = tags['TITLE']

        # add missing tags from cue sheet
        for (k, v) in self._cue_album_tags().items():
            if k not in tags:
                tags[k] = v

        return tags

    def _track_tags(self, num):
        t = self.track(num)
        tags = {}
        if t is not None:
            if t.artists:     tags['ARTIST'] = t.artists
            if t.composers:   tags['COMPOSER'] = t.composers
            if t.isrc:        tags['ISRC'] = [t.isrc]
            if t.num:         tags['TRACKNUMBER'] = [t.num]
            if t.title:       tags['TITLE'] = [t.title]

        return tags

    def _tag_as_flac_arg(self, tag_name, tag_value):
        log.debug(f'{tag_name}={tag_value}')
        # for whatever weird reason quote has a problem with some values if
        # not casted into str before
        return f"--tag={shlex.quote(str(tag_name))}={shlex.quote(str(tag_value))}"

    def track_tags_as_flac_args(self, num):
        tags = self._album_tags()
        for (k, vs) in self._track_tags(num).items():
            tags[k] = vs

        if 'TRACKTOTAL' not in tags:
            tags['TRACKTOTAL'] = [str(len(self.tracks()))]
        if 'ARTIST' in tags:
            tags['COMPOSER'] = tags['ARTIST']

        log.debug(f"tags for current track: {tags}")
        return ' '.join([self._tag_as_flac_arg(k, v) for k, vs in tags.items() for v in vs])


def init(ignore):
    log.info(f'Tags to ignore: "{ignore}"')
    AlbumInfo.IGNORE_TAGS_REX = re.compile(ignore)


@lru_cache(maxsize=5)
def get(path) -> AlbumInfo:
    return AlbumInfo(path)



#!/usr/bin/env python3
# 
# Copyright 2020-2021 by Andreas Schmidt
# All rights reserved.
# This file is part of the trackfs project
# and licensed under the terms of the GNU Lesser General Public License v3.0.
# See https://github.com/andresch/trackfs for details.
#

import os
import time

from dataclasses import dataclass
from tempfile import mkstemp
from subprocess import DEVNULL, run
from concurrent.futures import ThreadPoolExecutor
from typing import Tuple, Dict
from threading import RLock, Thread

from . import albuminfo
from .fusepath import FusePath
from .cuesheet import Track

import logging

log = logging.getLogger(__name__)


class FlacSplitException(Exception):
    pass


@dataclass(frozen=True)
class TrackInfo:
    temp_file_path: os.PathLike
    ref_count: int = 1
    last_accessed: float = time.time()


class TrackManager:
    """Keeps track of all individual tracks that currently get processed

    Each track has a unique key (usually the path of the virtual track file).

    The registry distinguishes three states for a track:
    * Unregistered: The track is not (yet) known to the registry
    * Announced: The track is known, but not yet available yet (processing still ongoing)
    * Available: The information about the track is available.

    """

    DEFAULT_TEMP_FILE_TTL = 60
    # We should keep the lead time big enough, as the calculation of the 
    # remaining track time is based on percentage of file-size
    DEFAULT_PRELOAD_LEAD_TIME = DEFAULT_TEMP_FILE_TTL // 2

    def __init__(self) -> None:
        self.rwlock: RLock = RLock()
        self.registry: Dict[os.PathLike, TrackInfo or None] = {}
        self.preload_pool: ThreadPoolExecutor = ThreadPoolExecutor(max_workers=2, thread_name_prefix="preload")
        self.preloaded_next_tracks: Dict[os.PathLike, FusePath] = {}
        self.preload_lead_time: int = TrackManager.DEFAULT_PRELOAD_LEAD_TIME
        self.temp_file_ttl: int = TrackManager.DEFAULT_TEMP_FILE_TTL

    def _add(self, key: os.PathLike, track_file: os.PathLike) -> None:
        with self.rwlock:
            self.registry[key] = TrackInfo(track_file)

        def cleanup():
            still_in_use = True
            while still_in_use:
                with self.rwlock:
                    info = self.registry[key]
                if info.ref_count <= 0 and (time.time() - info.last_accessed > self.temp_file_ttl):
                    still_in_use = False
                else:
                    time.sleep(self.temp_file_ttl / 2)
            log.debug(f'delete track "{key}"')
            del self.registry[key]
            os.remove(info.temp_file_path)

        Thread(target=cleanup).start()

    def _is_unregistered(self, key: os.PathLike) -> bool:
        """Is the track at the given key not yet registered?"""
        with self.rwlock:
            return key not in self.registry

    def _announce(self, key: os.PathLike) -> None:
        """Announce that a new track will get processed soon"""
        with self.rwlock:
            self.registry[key] = None

    def _is_announced(self, key: os.PathLike) -> bool:
        """Is the track registered, but not yet processed?"""
        with self.rwlock:
            # default value "" for get ensures that we don't 
            # treat unknown tracks as announced
            return self.registry.get(key, "") is None

    def _is_registered(self, key: os.PathLike) -> bool:
        with self.rwlock:
            return isinstance(self.registry.get(key), TrackInfo)

    def _change_usage(self, key: os.PathLike, delta: int) -> TrackInfo:
        with self.rwlock:
            info = self.registry[key]
            info = TrackInfo(info.temp_file_path, info.ref_count + delta)
            self.registry[key] = info
        return info

    def __getitem__(self, key: os.PathLike) -> TrackInfo:
        with self.rwlock:
            return self.registry[key]

    def get(self, key: os.PathLike, default: TrackInfo = None) -> TrackInfo:
        with self.rwlock:
            return self.registry.get(key, default)

    def __delitem__(self, key: os.PathLike) -> None:
        with self.rwlock:
            del self.registry[key]

    @staticmethod
    def _new_temp_filename() -> os.PathLike:
        (fh, temp_file) = mkstemp()
        # we don't want to process the file in python; just want a unique filename
        # that we let flac write the track into
        # => close right away
        os.close(fh)
        return temp_file

    def _extract_track(self, path: os.PathLike, fp: FusePath) -> os.PathLike:
        """creates a real file for a given virtual track file

        extracts the track from the underlying FLAC+CUE file into 
        a temporary file and then opens the temporary file"""
        log.info(f'open track "{path}"')

        track_file = self._new_temp_filename()

        flac_info = albuminfo.get(fp.source)

        # extract picture from flac if available
        picture_file = self._new_temp_filename()
        metaflac_cmd = f'metaflac --export-picture-to="{picture_file}" "{fp.source}"'
        log.debug(f'extracting picture with command: "{metaflac_cmd}"')
        rc = run(metaflac_cmd, shell=True, stdout=None, stderr=DEVNULL).returncode
        picture_arg = ""
        if rc == 0:
            picture_arg = f' --picture="{picture_file}"'

        flac_cmd = (
            f'flac -d --silent --stdout --skip={fp.start.flac_time()}'
            f'  --until={fp.end.flac_time()} "{fp.source}" '
            f'| flac --silent -f --fast'
            f'  {flac_info.track_tags_as_flac_args(fp.num)}{picture_arg} -o {track_file} -'
        )
        log.debug(f'extracting track with command: "{flac_cmd}"')
        rc = run(flac_cmd, shell=True, stdout=None, stderr=DEVNULL).returncode
        os.remove(picture_file)
        with self.rwlock:
            if rc != 0:
                err_msg = f'failed to extract track #{fp.num} from file "{fp.source}"'
                log.error(err_msg)
                os.remove(track_file)
                del self[path]
                raise FlacSplitException(err_msg)
            else:
                self._add(path, track_file)

        return track_file

    def prepare_track(self, path: os.PathLike, fp: FusePath) -> os.PathLike:
        log.info(f'prepare track "{path}"')
        assert fp.is_track
        ready_to_process = False
        while not ready_to_process:
            with self.rwlock:
                if self._is_unregistered(path):
                    ready_to_process = True
                    self._announce(path)
                elif self._is_registered(path):
                    # we already have cached that track => 
                    # register additional usage
                    return self._change_usage(path, +1).temp_file_path

            # give other thread time to finish processing
            time.sleep(0.5)
        return self._extract_track(path, fp)

    def release_track(self, path: os.PathLike, fp: FusePath):
        log.info(f'release track "{path}"')
        assert fp.is_track
        if self._change_usage(path, -1).ref_count == 0:
            log.debug(f'check fo preloaded next track of "{path}"')
            with self.rwlock:
                next_track = self.preloaded_next_tracks.get(path, None)
                if next_track is not None:
                    log.debug(f'release preloaded next track of "{next_track.vpath}"')
                    self.release_track(next_track.vpath, next_track)
                    del self.preloaded_next_tracks[path]

    @staticmethod
    def _find_this_and_next_track(flac_info: albuminfo.AlbumInfo, num: int) -> Tuple[Track or None, Track or None]:
        log.info(f'checking for subsequent track of track "{num}"')
        tracks = flac_info.tracks()
        if tracks is not None:
            total_tracks = len(tracks)
            track = None
            found = False
            i = num - 1
            while (not found) and (i < total_tracks):
                track = tracks[i]
                found = track.num == num
                i += 1
            return track, tracks[i] if i < total_tracks else None
        else:
            log.warning('could not find any tracks')
            return None, None

    def _do_check_next_track(self, path: os.PathLike, fp: FusePath, offset: int) -> None:
        log.info(f'_do_check_next_track: "{path}" [{offset}]')

        duration = (fp.end - fp.start).seconds()
        file_size = os.stat(self[path].temp_file_path).st_size
        if (1.0 - (float(offset) / float(file_size))) * duration > self.preload_lead_time:
            log.debug(f'more than ~{self.preload_lead_time} seconds to play; no preload')
            return

        flac_info = albuminfo.get(fp.source)
        (track, next_track) = self._find_this_and_next_track(flac_info, fp.num)
        if next_track is None:
            log.debug(f'got last track: "{fp.num}"; no preload')
            return

        with self.rwlock:
            if self.preloaded_next_tracks.get(path, None) is not None:
                log.debug(f'next track of "{path}" got preloaded in the meanwhile')
                return
            # mark as preloaded, so that we don't do it twice

            log.debug(f'preloading next track "{next_track.num}"')
            next_fp = fp.for_other_track(
                next_track.num, next_track.title, next_track.start, next_track.end,
            )
            self.preloaded_next_tracks[path] = next_fp

        # prepare next track and by that add a reference to it to keep it it cache
        # for the lifetime if this track.
        self.prepare_track(next_fp.vpath, next_fp)

    def check_next_track(self, path: os.PathLike, fp: FusePath, offset: int) -> None:
        assert fp.is_track
        log.info(f'check_next_track for "{path}" [{offset}]')
        with self.rwlock:
            if self.preloaded_next_tracks.get(path, None) is not None:
                log.debug(f'next track already preloaded "{path}"')
                return
        log.info(f'enqueue next track check of "{path}" [{offset}]')
        self.preload_pool.submit(self._do_check_next_track, path, fp, offset)

    def estimate_track_file_size(self, path: os.PathLike, fp: FusePath) -> int:
        track_info = self.get(path, None)
        if track_info is None:
            # TODO: can we find a better estimation?
            # use raw-audio size as estimation
            f = albuminfo.get(fp.source).meta
            return int(
                (fp.end - fp.start).seconds()
                * f.info.channels
                * (f.info.bits_per_sample / 8)
                * f.info.sample_rate
            )
        else:
            # use the actual size of the track-file
            return os.stat(track_info.temp_file_path).st_size
